library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.ALL;

-- this module defines a simple digital audio interface from 3-wire i2s format to 17-bit 
-- L/R slot multiplexed parallel data
-- The 17-bit data is rectified and 16-bit slot-multiplexed data is output at the start of
-- bitslot 19 in each L or R subframe

entity i2s_iface is
    port (
	RESETL: 		in STD_LOGIC;
	CLK64fs: 		in STD_LOGIC;
	CLKfs: 			in STD_LOGIC;
	SDATAfs: 		in STD_LOGIC;
	LSLOT: 			out STD_LOGIC_VECTOR (4 downto 0);
	RSLOT:			out STD_LOGIC_VECTOR (4 downto 0);
	LR_RECTDATA: 	out STD_LOGIC_VECTOR (15 downto 0)
	);
	
end i2s_iface;

architecture i2s_iface_arch of i2s_iface is
	signal LRshiftreg: STD_LOGIC_VECTOR(16 downto 0);  -- note 17 bits = 16 bits + S
	signal LRbufreg: STD_LOGIC_VECTOR(15 downto 0);
	signal Lcount, Rcount: STD_LOGIC_VECTOR(4 downto 0);
	signal SREN: STD_LOGIC;

begin
	-- set up bitcounters synchronous with CLKfs
	process(CLK64fs,RESETL) begin
		if RESETL='0' then
			Lcount <=(others=>'0');
			Rcount <=(others=>'0');
		elsif CLK64fs'event and CLK64fs='1' then
			if CLKfs='1' then
				Lcount <=(others=>'0');		-- right counter counts when CLKfs HI
				Rcount <= Rcount+1;		
			else
				Lcount <=Lcount+1;			-- left counter counts when CLKfs LO
				Rcount <=(others=>'0');
			end if;
		end if;		
	end process;
	
	-- define slotcount outputs
	LSLOT <= Lcount;
	RSLOT <= Rcount;
	
	-- define control signals
	SREN <='1' when ((Lcount >= 1 and Lcount <= 17) or (Rcount >= 1 and Rcount <= 17)) else '0';

	-- define shift register 17 bits long, enabled during 17 bit frames of each L and R channel
	
	process(CLK64fs, RESETL) begin
	if RESETL='0' then
		LRshiftreg <=(others=>'0');
	elsif CLK64fs'event and CLK64fs='1' then
		if SREN='1' then
			LRshiftreg(16 downto 1) <= LRshiftreg(15 downto 0);
			LRshiftreg(0) <= SDATAfs;
		else
			LRshiftreg(16 downto 0) <= LRshiftreg(16 downto 0);
		end if;
	end if;	
	end process;	

		
	-- rectify shiftreg data and clock into mux'd buffer reg at start of slot 21L and 21R
	process(CLK64fs,RESETL) begin
	if RESETL = '0' then
		LRbufreg <=(others=>'0');
	elsif CLK64fs'event and CLK64fs='1' then
		if ((Lcount=19) or (Rcount=19)) then
			if LRshiftreg(16)='1' then
				LRbufreg(15 downto 0) <= not LRshiftreg(15 downto 0);	
			else
				LRbufreg(15 downto 0) <= LRshiftreg(15 downto 0);
			end if;
		else
			LRbufreg <= LRbufreg;
		end if;
	end if;	
	end process;		
	
	LR_RECTDATA <= LRbufreg;
		
end i2s_iface_arch;
