library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.ALL;

-- This module implements a 21-LED combined PPM/VU bargraph level display on a
-- 16-bit rectified L/R slot multiplexed audio input word RECTWORDSLR.  
-- A Bar of 10 Green LEDs covers the interval -66dB to -12dB. 
-- Spacing is 6dB determined by simple bit weighting
-- A Bar of Orange LEDs covers the interval -10dB to -1dB. 
-- Spacing is 1dB determined by 8-bit MSBs magn. comp
-- A Single Red LED at the top of the column indicates clipping point 0dB 
-- Functional flow consists of bar encoding, a peak-hold detector and a decay mechanism
-- that decays peak at the rate of DECAYCLK.

-- Rev B 21/10/2001: 
-- ***  A simple peak-following resampler is added to the solid column to reduce
--     	refresh rate and flicker
-- ***  Start of PeakHold refresh hold time synchronised with occurrence of new peak.
-- ***  PeakHold made to go off instead of slow decay. 

-- Rev C 14/08/2002:
-- *** 'Orange Region' extended to be -10dB to -1dB.
-- *** 'Green Region'  changed from -12dB to -90dB
-- *** Total number of LEDs reduced to 21  (was 22)

entity vu_ppm1 is
    port 
    	(
		LR_RECTWORDS: in STD_LOGIC_VECTOR (15 downto 0) ;
		INREGCLKEN: in STD_LOGIC;
		PKREGCLKEN: in STD_LOGIC;
		AVGCLKEN: in STD_LOGIC;
		CLK64fs: in STD_LOGIC;
		RESETL: in STD_LOGIC;
		BARLED: out STD_LOGIC_VECTOR (20 downto 0) 
		);
-- note:  PKREGCLKEN and DECAYCLKEN must be simultaneous and delayed after INREGCLKEN
--        PKREGCLKEN and AVGCLKEN must be simultaneous and delayed after INREGCLKEN
end vu_ppm1;

architecture vu_ppm1_arch of vu_ppm1 is
	-- HOLDTIME determines how long the peak LED remains static
	constant HOLDTIME	: STD_LOGIC_VECTOR (6 downto 0):= B"1111111";   
	signal  HoldCntr	: STD_LOGIC_VECTOR (6 downto 0);
	signal  InputReg	: STD_LOGIC_VECTOR (15 downto 0);
	signal  SolidBar,PeakBar,DecaySR,PeakReg,SolidReg1,SolidReg2
						: STD_LOGIC_VECTOR (20 downto 0);
	signal  NewPeak,HoldOver: STD_LOGIC;
	type LUT is array (0 to 10) of STD_LOGIC_VECTOR (7 downto 0);
	constant Thresholds	: LUT:= LUT'(	X"51",		-- -10dB
													X"5A",		-- -9dB
													X"65",		-- -8dB
													X"71",		-- -7dB
													X"7F",		-- -6dB
													X"8F",		-- -5dB
									 				X"A0", 		-- -4dB
									 				X"B4",		-- -3dB
									 				X"CA",		-- -2dB
									 				X"E3",		-- -1dB
									 				X"FD");		--  0dB allowing for mul rounding error	

begin

-- Clock the audio words from the multiplexed L/R bus into the input register
	process(CLK64fs,RESETL) begin
		if RESETL='0' then 
			InputReg <=(others=>'0');
		elsif CLK64fs'event and CLK64fs='1' then
			if INREGCLKEN='1' then
				InputReg <= LR_RECTWORDS;	
			end if;
		end if;
	end process;

-- convert input audio words to a 'solid' barcode and 
-- Determine the peak value as a 'floating' single barled

-- First for the ORANGELED Region using magnitude comparison. . .
	process(InputReg, SolidBar) begin
		if InputReg(15 downto 8) >= Thresholds(10) then		-- 1 x Red LED 0dBFS 
			SolidBar(20) <= '1';
			PeakBar(20) <= '1';
		else
			SolidBar(20) <= '0';
			PeakBar(20) <= '0';
		end if;	
		for i in 10 to 19 loop										-- 10 x Orange LEDS -1dBFS..-10dBFS
			if InputReg(15 downto 8) >= Thresholds(i-10) then
				SolidBar(i) <= '1';
				PeakBar(i) <= not SolidBar(i+1) and SolidBar(i);
			else
				SolidBar(i) <= '0';	
				PeakBar(i) <= '0';
			end if;
		end loop;
	end process;
		
-- Second for the GREENLED Region using bit weighting. . .
	process(InputReg, SolidBar) begin
		for i in 0 to 9 loop											
			SolidBar(i) <= SolidBar(i+1) or InputReg(i+5);    	-- InputReg(14) => -12dBFS
			PeakBar(i) <= not SolidBar(i+1) and SolidBar(i);	
		end loop;
	end process;	


-- Create 2-deep registered stack SolidBar1,2, enabled with PKREGCLKEN and AVGCLKEN.
-- This is a resampling peak-follower (RPF) at freq AVGCLKEN

	process(CLK64fs,RESETL) begin
		if RESETL='0' then 
			SolidReg1 <=(others=>'0');
			SolidReg2 <=(others=>'0');
		elsif CLK64fs'event and CLK64fs='1' then
			if PKREGCLKEN='1' then
				if AVGCLKEN = '1' then
					SolidReg1 <= SolidBar;					-- Reset PF to new input bar at resample interval
					SolidReg2 <= SolidReg1;					-- Transfer RPF to output
				else
					SolidReg1 <= SolidBar or SolidReg1;		-- Update PF on every input bar
					SolidReg2 <= SolidReg2;		
				end if;
			end if;
		end if;
	end process;

-- Implement the HoldCntr, which is used to time how long the 'fresh' peak is held.
-- The counter is enabled by AVGCLKEN and reloaded with HOLDVAL every time a 'fresh'
-- peak occurs.  In the absense of 'fresh peaks' the counter counts down and when it
-- reaches 0 the peak hold is cleared and a new peak value loaded.

	process(CLK64fs,RESETL) begin
		if RESETL='0' then 
			HoldCntr <=(others=>'0');
		elsif CLK64fs'event and CLK64fs='1' then
			if PKREGCLKEN='1' then
				if NewPeak='1' then					-- Load HOLDTIME on next clock...
					HoldCntr <= HOLDTIME;
				elsif AVGCLKEN='1' then
					HoldCntr <= HoldCntr - 1;		-- ...or dec counter only on next AVGCLKEN
				end if;
			end if;
		end if;
	end process;

-- Intermediate var defines when the hold time is done 
	HoldOver <= '1' when HoldCntr = 0 else '0';

-- Compare registered PeakBar with new data to determine if new peak
	NewPeak <= '1' when PeakBar > PeakReg else '0';
			
-- Implement Peak register; loaded on next clock with new data if PeakBar is higher than
-- existing (registered) value or HoldCntr reaches zero, otherwise the value is held. 
-- All clock ops are enabled with PKREGCLKEN.

	process(CLK64fs,RESETL) begin
		if RESETL='0' then PeakReg <=(others=>'0');
		elsif CLK64fs'event and CLK64fs='1' then
			if PKREGCLKEN='1' then
				if NewPeak='1' or HoldOver='1' then
					PeakReg <= PeakBar;
				end if; 	
			end if;
		end if;
	end process;

-- Merge Decay controlled floating peak with the solid bar and assign activ hi outputs
-- with reset blanking
	BARLED(20 downto 0) <= (PeakReg or SolidReg2) when RESETL='1' else (others=>'1');

end vu_ppm1_arch;
